/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Timo Weberruß ( timo dot weberruss at student dot kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include "DistanceFunction.h"
#include "GroupShapeGenerator.h"
#include "armarx/navigation/core/basic_types.h"
#include "armarx/navigation/human/types.h"

namespace armarx::navigation::human
{
    /**
     * @brief Identifies social interaction groups in a collection of detected humans.
     *
     * Can identify social interaction groups in a collection of humans. The currently detected
     * humans can be set by calling updateHumans.
     */
    class HumanGrouper
    {

        using Human = armarx::navigation::human::Human;

    public:
        /**
         * @brief The GroupingSettings struct contains settings for group generation.
         * This includes the influence of aligned orientation and how aggressively groups
         * are generated.
         */
        struct GroupingSettings
        {
            const double groupingThreshold;
            const double maxOrientationInfluence;
            const double movementInfluence;
        };

        /**
         * @brief Creates a new HumanGrouper
         * @param settings different parameters that will affect group generation
         */
        HumanGrouper(const GroupingSettings& settings);

        /**
         * @brief Sets the currently detected humans.
         *
         * Sets the currently detected humans. Only these humans will be assumed to currently exist
         * when getCurrentGroups() is called the next time.
         * @param newHumans a vector containing the most recently detected humans
         */
        void updateHumans(const std::vector<Human>& newHumans);

        /**
         * @brief Recognizes groups in the current humans.
         *
         * Identifies and returns social groups in the humans given to this instance by the last
         * call to updateHumans.
         * @return a vector containing the recognized human groups
         */
        std::vector<human::HumanGroup> getCurrentGroups();


        void setCurrentHumans(const std::vector<Human>& newCurrentHumans);

    private:
        const std::unique_ptr<DistanceFunction> distanceFunction;
        const std::unique_ptr<GroupShapeGenerator> shapeGenerator;
        const GroupingSettings settings;

        std::vector<Human> currentHumans_;
    };
} // namespace armarx::navigation::human
