/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Timo Weberruß ( timo dot weberruss at student dot kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include "DistanceFunction.h"


namespace armarx::navigation::human
{
    /**
     * @brief The OrientationDistance class is a DistanceFunction that computes the distance between
     * the two humans based upon their orientation towards or away from each other.
     *
     * This distance function will return
     * - the specified minimum distance (0 by default) if the humans are exactly facing each other
     * - the specified maximum distance (1 by default) if the humans are oriented back to back
     * - a linear interpolation between the two for anything inbetween (for instance, two people
     *   looking in the same direction will have a distance equal to the arithmetic mean of the
     *   minimum and maximum distance, 0.5 by default)
     */
    class OrientationDistance : public DistanceFunction
    {
    public:
        /**
         * @brief OrientationDistance Creates a new OrientationDistance with the specified range
         * @param max the distance two people will have when faced back to back
         * @param min the distance two people will have when oriented face to face
         */
        OrientationDistance(double max = 1, double min = 0);
        double computeDistance(const Human& h1, const Human& h2) const override;

    private:
        const double max;
        const double min;

        // helper methods
        static double getOrientationFactor(const Human& h1, const Human& h2);
        static double getLineOrientation(const Human& h2, const Human& h1);
        static double normOrientation(double orientation);
    };
} // namespace armarx::navigation::human
