#include "TimedElasticBands.h"
#include <optional>

#include <SimoxUtility/algorithm/apply.hpp>
#include <SimoxUtility/json/json.hpp>
#include <VirtualRobot/MathTools.h>
#include <VirtualRobot/Robot.h>
#include <VirtualRobot/SceneObjectSet.h>

#include <ArmarXCore/core/PackagePath.h>
#include <ArmarXCore/core/logging/Logging.h>

#include "RobotAPI/libraries/aron/core/data/rw/reader/nlohmannJSON/NlohmannJSONReaderWithoutTypeCheck.h"

#include "armarx/navigation/conversions/eigen.h"
#include "armarx/navigation/core/Trajectory.h"
#include <armarx/navigation/local_planning/LocalPlanner.h>
#include <armarx/navigation/local_planning/aron/TimedElasticBands.aron.generated.h>
#include <armarx/navigation/local_planning/aron_conversions.h>
#include <armarx/navigation/local_planning/core.h>
#include <armarx/navigation/local_planning/ros_conversions.h>
// #include <teb_local_planner/extension/obstacles/EllipseObstacle.h>
#include <teb_local_planner/homotopy_class_planner.h>

namespace armarx::navigation::local_planning
{

    // TimedElasticBandsParams

    Algorithms
    TimedElasticBandsParams::algorithm() const
    {
        return Algorithms::TimedElasticBands;
    }

    aron::data::DictPtr
    TimedElasticBandsParams::toAron() const
    {
        return cfg.toAron();
    }

    TimedElasticBandsParams
    TimedElasticBandsParams::FromAron(const aron::data::DictPtr& dict)
    {
        ARMARX_CHECK_NOT_NULL(dict);

        TimedElasticBandsParams bo;
        bo.cfg.fromAron(dict);

        return bo;
    }


    // TimedElasticBands

    TimedElasticBands::TimedElasticBands(const Params& i_params, const core::Scene& ctx) :
        LocalPlanner(ctx), params(i_params), scene(ctx), obstManager(teb_obstacles)
    {
        //TODO (SALt): find proper place to init with default config
        if (true)
        {
            readDefaultConfig(params.cfg);
        }
        toTebCfg(params.cfg, cfg_);

        auto robot_model = boost::make_shared<teb_local_planner::CircularRobotFootprint>(
            params.cfg.robot_footprint_radius);

        hcp_ = std::make_unique<teb_local_planner::HomotopyClassPlanner>();
        hcp_->initialize(
            cfg_, &teb_obstacles, robot_model, teb_local_planner::TebVisualizationPtr(), nullptr);

        ros::Time::init(); // we have to init time before we can use the planner
    }

    void
    TimedElasticBands::readDefaultConfig(arondto::TimedElasticBandsParams& target)
    {
        const armarx::PackagePath configPath("armarx_navigation",
                                             "local_planner_config/TimedElasticBands/default.json");
        const std::filesystem::path file = configPath.toSystemPath();

        ARMARX_CHECK(std::filesystem::is_regular_file(file)) << file;

        ARMARX_INFO << "Loading config from file `" << file << "`.";
        std::ifstream ifs{file};

        nlohmann::json jsonConfig;
        ifs >> jsonConfig;

        ARMARX_INFO << "Reading config";

        armarx::aron::data::reader::NlohmannJSONReaderWithoutTypeCheck reader;
        target.read(reader, jsonConfig);
    }

    std::optional<LocalPlannerResult>
    TimedElasticBands::plan(const core::GlobalTrajectory& goal)
    {
        const core::Pose currentPose{scene.robot->getGlobalPose()};

        // prune global trajectory
        const auto& [prunedGoal, planToDest] = goal.getSubTrajectory(
            currentPose.translation(), params.cfg.planning_distance * 1000); // [m] to [mm]

        const teb_local_planner::TimedElasticBand globalPath = conv::toRos(prunedGoal);
        teb_globalPath = globalPath.poses();
        hcp_->setGlobalPath(&teb_globalPath);

        const teb_local_planner::PoseSE2 start = conv::toRos(currentPose);
        const teb_local_planner::PoseSE2 end =
            conv::toRos(prunedGoal.points().back().waypoint.pose);

        geometry_msgs::Twist velocity_start = conv::toRos(scene.platformVelocity);


        fillObstacles();

        try
        {
            hcp_->plan(start, end, &velocity_start, !planToDest);
        }
        catch (std::exception& e)
        {
            ARMARX_DEBUG << "Caugth exception while planning: " << e.what();
            return {};
        }

        if(hcp_->getTrajectoryContainer().empty())
        {
            ARMARX_VERBOSE << "Did not find any trajectory!";
            return std::nullopt;
        }

        ARMARX_VERBOSE << "Planned successfully (found " << hcp_->getTrajectoryContainer().size()
                       << " Trajectories)";

        core::LocalTrajectory best = conv::fromRos(hcp_->findBestTeb()->teb());

        // visualize path alternatives
        if (arviz)
        {
            auto layer = arviz.value().layer("local_planner_path_alternatives");

            int i = 0;
            int bestTebIdx = hcp_->bestTebIdx();
            for (const auto& teb : hcp_->getTrajectoryContainer())
            {
                if (i == bestTebIdx)
                {
                    continue;
                }
                const core::LocalTrajectory trajectory = conv::fromRos(teb->teb());
                const std::vector<Eigen::Vector3f> points =
                    simox::alg::apply(trajectory.points(),
                                      [](const core::LocalTrajectoryPoint& pt) -> Eigen::Vector3f
                                      { return pt.pose.translation(); });
                layer.add(viz::Path("path_alternative_" + std::to_string(i))
                              .points(points)
                              .color(simox::Color::gray()));
                i++;
            }

            arviz.value().commit(layer);
        }

        return {{.trajectory = best}};
    }

    void
    TimedElasticBands::fillObstacles()
    {
        obstManager.clear();

        viz::Layer* visPtr = nullptr;
        viz::Layer visLayer;
        if (arviz)
        {
            visLayer = arviz.value().layer("local_planner_obstacles");
            visPtr = &visLayer;
        }


        if (scene.staticScene)
        {
            ARMARX_CHECK(scene.staticScene.has_value());
            for (const auto& obst : scene.staticScene.value().objects->getCollisionModels())
            {
                obstManager.addBoxObstacle(obst->getGlobalBoundingBox(), visPtr);
            }
        }
        if (scene.dynamicScene)
        {
            ARMARX_CHECK(scene.dynamicScene.has_value());
            for (const auto& obst : scene.dynamicScene.value().humans)
            {
                obstManager.addHumanObstacle(obst, visPtr);
            }
        }

        if (arviz)
        {
            arviz.value().commit(visLayer);
        }

        ARMARX_INFO << "TEB: added " << obstManager.size() << " obstacles";
    }

} // namespace armarx::navigation::local_planning
